# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek

import output
from scheduler.i18n import _

import smtplib
import datetime

from email.MIMEText import MIMEText 
from email.MIMEMultipart import MIMEMultipart
from subprocess import Popen, PIPE
from ConfigParser import NoOptionError 

class MailOutput(object):
    """
    Allows composing ST_TestResult and ST_TestResultDiff objects into an email
    and sending it via the method specified in the config file.
    """

    def __init__(self, options, schedopts, from_addr='sectool', to_addr=None, subject=None, formatter=output.StdoutFormatter()):
        self.options = options
        self.mail_options = schedopts['MAIL']
        self.payload = ""
        self.attachment = ""
        self.att_name = "sectool-report-" + datetime.date.today().isoformat() + ".txt"
        self.msg = None
        self.formatter = formatter

        if to_addr == None:
            to_addr = self.options.mail

        if subject == None:
            subject = "sectool report " + datetime.date.today().isoformat() 

        self.msg = MIMEMultipart()
        self.msg['From'] = from_addr
        self.msg['To']   = to_addr
        self.msg['Subject'] = subject


    def __str__(self):
        if self.msg == None:
            return ''
        else:
            return self.msg.as_string()

    def _add(self, content, to_body):
        """ 
        Adds content either to body or to attachment 

        @param content: the content to add
        @type content: string
        @param to_body: true if add to body, false if add to attachment
        @type to_body: boolean
        """
        if to_body:
            self.payload += content
        else:
            self.attachment += content

    def add_result(self, test_name, result, to_body=True):
        """
        Adds test result to self

        @param test_name: name of the test to add
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param to_body: true if add to body, false if add to attachment
        @type to_body: boolean
        """
        self._add(self.formatter.test_result_header(test_name) + '\n', to_body)
        self._add(self.formatter.output_one_result(result, self.options), to_body)
        self._add(self.formatter.test_result_footer(test_name, result.result) + '\n', to_body)
        
    def add_diff(self, test_name, result, diff, to_body=True):
        """
        Adds test result with a diff to self

        @param test_name: name of the test to add
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param result: diff of result and old result
        @type result: ST_TestResultDiff
        @param to_body: true if add to body, false if add to attachment
        @type to_body: boolean
        """
        self._add(self.formatter.output_one_diff(test_name, diff, self.options) + '\n', to_body)
        if diff and diff.are_same == False:
            self._add(self.formatter.diff_result_footer(test_name, diff) + '\n', to_body)

    def _construct_attachment(self):
        """
        Creates an actual attachment named self.att_name from the string
        located at self.attachment
        """
        att = None

        if self.attachment != "":
            att = MIMEText(self.attachment)
            att.add_header('Content-Disposition', 'attachment', filename=self.att_name)

        return att

    def fini(self):
        """
        Creates the actual e-mail message by adding self.payload as the body
        and self.attachment as attachment
        """
        self.msg.attach(MIMEText(self.payload))
        att = self._construct_attachment()
        if att:
            self.msg.attach(att)

    def _send_local(self):
        """
        Sends the message composed with self.fini() using local sendmail                
        """

        sendmail = Popen('/usr/sbin/sendmail -i -t', shell=True, stdin=PIPE, close_fds=True)

        sendmail.stdin.write(str(self))
        sendmail.stdin.close()

        return sendmail.wait()

    def _send_smtp(self):
        """
        Sends the message composed with self.fini() using a smtp server
        """

        smtp_server = smtplib.SMTP(self.mail_options['SMTP_SERVER'])

        try:
            user = self.mail_options['SMTP_USER']
        except NoOptionError: user = None
        try:
            passwd = self.mail_options['SMTP_PASSWD']
        except NoOptionError: passwd = None

        if user != None: 
            smtp_server.login(user, passwd)

        smtp_server.sendmail(self.msg['From'], self.msg['To'], str(self))
        smtp_server.quit()

        return 0

    def send(self):
        """
        Sends the message composed with self.fini() using the method
        specified in the config file.
        """

        target = self.mail_options['TARGET'].upper() 
        retval = 0

        if target == 'LOCAL':
            retval = self._send_local()
        elif target == 'SMTP':
            retval = self._send_smtp()
        else:
            raise ValueError(_('Neither TARGET=LOCAL nor TARGET=SMTP found in config file (section[MAIL]'))

        return retval


