# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek
# -*- coding: UTF8 -*-

from UserDict import UserDict

from scheduler.i18n import _
from scheduler.scheduler import OutputItem
from scheduler.scheduler import (ST_TestCase, ST_TestResult)

class RunReport(UserDict):
    def __init__(self):
        UserDict.__init__(self)
        self.total  = 0
        self[ST_TestResult.r_passed]     = 0
        self[ST_TestResult.r_warning]    = 0
        self[ST_TestResult.r_error]      = 0
        self[ST_TestResult.r_failure]    = 0
        self[ST_TestResult.r_fatal]      = 0


class OutputFormatter(object):
    """ A base class, meant to be subclassed. Defines several output methods, 
    one for every secTool action. """

    def output_one_diff(self, test_name, diff, options):
        """ 
        Outputs one test result diff

        @param test_name: name of the test
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param diff: diff against previous run
        @type diff: ST_TestResultDiff
        @param options: tui options
        """
        pass

    def output_one_result(self, results, options):
        """
        Outputs one test result

        @param test_name: name of the test
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param options: tui options
        """
        pass

    def output_test_list(self, tests):
        """ 
        Outputs a list of all known tests

        @param tests: the tests to list
        @type tests: ST_TestResult list
        """
        pass

    def output_test_info(self, tests):
        """ 
        Outputs info about a test 
        @param tests: the tests to list
        @type tests: ST_TestResult list
        """
        pass

    def output_summary(self, total, passed, warn, error, fail, fatal):
        """
        Outputs nice summary about all the tests that were run
        """
        pass

    def result_callback(self, item, cb_data):
        """
        This is a method that gets called directly from the scheduler after
        a message has been processed to allow real-time printing of items.

        @param item: the item to print
        @type item: scheduler.OutputItem
        @param cb_data: arbitrary data passed to the callback function. This
        must have been registered during register_callback
        @type cb_data: object
        """
        pass

    def test_result_header(self, test_name):
        """
        Outputs a nicely formatted header for one test

        @param test_name: name of the test
        @type test_name: string
        """
        pass

    def test_result_footer(self, test_name, test_result):
        """
        Outputs a nicely formatted header for one test

        @param test_name: name of the test
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        """
        pass

    def diff_result_footer(self, test_name, diff):
        """
        Outputs a nicely formatted footer for one test

        @param test_name: name of the test
        @type test_name: string
        @param diff: diff against previous run
        @type diff: ST_TestResultDiff
        """
        pass

class StdoutFormatter(OutputFormatter):
    """ Output which typically aims to a tty and will be read by a human. This
    is also secTool default. """

    # padding for listing results
    w_total = 80

    # padding for listing tests
    list_w_name   = 15
    list_w_levels = 10 
    list_w_desc = w_total - list_w_name - list_w_levels

    # padding for info results
    info_pad = 12
    report_pad = 15

    str_test_results = [ _("PASS"), _("WARNING"), _("ERROR"), _("FAIL"), _("FATAL")]
    color_test_results = [ 'green', 'purple', 'red', '', '' ]
    str_labels = { OutputItem.t_error : _("Error"),
                   OutputItem.t_warning : _("Warning"),
                   OutputItem.t_hint : _("Hint"),
                   OutputItem.t_info : _("Info") }

    colors = {  '':'', "black":"30", "red":"31", "green":"32", "brown":"33", "blue":"34",
    "purple":"35", "cyan":"36", "lgray":"37", "gray":"1;30", "lred":"1;31",
    "lgreen":"1;32", "yellow":"1;33", "lblue":"1;34", "pink":"1;35",
    "lcyan":"1;36", "white":"1;37" }

    def _textcolor(self, text, col="", raw=False):
        if raw:
            return text
        return "\033[%sm%s\033[0m" % (self.colors[col], text)

    def _textbold(self, text, raw=False):
        if raw:
            return text
        return "\033[1m%s\033[0;0m" % text

    def output_one_diff(self, test_name, diff, options, raw = True):
        """
        See superclass definition

        @rtype: string
        """
        payload = ""

        if diff.are_same == True:
            if options.debug:
                payload += _("%s\n\tThe test did not change its result") % self.test_result_header(test_name, raw)
            return payload

        payload += self.test_result_header(test_name, raw)
        for item in diff.added:
            msg = ("\n\t + %s(%s)\t" % (self.str_labels[item.type], item.msg_id)).ljust(15)
            msg += item.msg.lstrip()
            payload += "%s" % msg

        for item in diff.removed:
            msg = ("\n\t - %s(%s)\t" % (self.str_labels[item.type], item.msg_id)).ljust(15)
            msg += item.msg.lstrip()
            payload += "%s" % msg

        return payload

    def diff_result_footer(self, test_name, diff, raw = True):
        """
        Returns a nicely formatted diff footer

        @param test_name: name of the test
        @type test_name: string
        @param diff: diff against previous run
        @type diff: ST_TestResultDiff
        @param options: tui options
        @param raw: whether to ommit special formatting (terminal control characters)
        @type raw: boolean

        @rtype: string
        """
        if diff.are_same or (diff.before.result == diff.after.result):
            return _("\t%s:%s\n") % (self._textbold(test_name, raw),
                                     self._textcolor(self.str_test_results[diff.before.result],
                                                     self.color_test_results[diff.before.result],
                                                     raw))
        else:
            return _("\t%s:%s (Changed from: %s)\n") % (self._textbold(test_name, raw),
                                                        self._textcolor(self.str_test_results[diff.after.result],
                                                                        self.color_test_results[diff.after.result],
                                                                        raw),
                                                        self._textcolor(self.str_test_results[diff.before.result],
                                                                        self.color_test_results[diff.before.result],
                                                                        raw))


    def test_result_header(self, test_name, raw=True):
        """
        Returns a nicely formatted test header

        @param test_name: name of the test
        @type test_name: string
        """
        return _("%s ->") % (self._textbold(test_name, raw))

    def test_result_footer(self, test_name, test_result, raw=True):
        """
        Returns a nicely formatted test footer

        @param test_name: name of the test
        @type test_name: string
        """
        return _("    %s: %s\n") % (self._textbold(test_name, raw),
                                   self._textbold(self.str_test_results[test_result], raw))
                                   #self._textcolor(self.str_test_results[test_result], 
                                   #                self.color_test_results[test_result],
                                   #                raw))

    def __output_one_item(self, item, options):
        """
        Formats a message about one output item and returns the string.

        @param item: the item to format
        @type item: scheduler.OutputItem
        @param options: options that tune the output 
        @type options: sectool.TuiOptions
        """
        # don't print hints or infos without the proper option
        if item.type == OutputItem.t_hint and options.hint == False:
            return None
        if item.type == OutputItem.t_info and options.debug == False:
            return None

        return "    %s: %s" % (self.str_labels[item.type], item.msg.strip())

    def result_callback(self, item, cb_data):
        """
        The actual callback called from the scheduler
        """
        output = self.__output_one_item(item, cb_data)
        if(output):
            print output

    def output_one_result(self, result, options):
        """
        See superclass definition

        @rtype: string
        """

        payload = ""

        # output stderr if the test ended up with a failure or if the user specified the --debug option
        if result.result == ST_TestResult.r_fatal or result.result == ST_TestResult.r_failure or options.debug:
            for failure in result.stderr:
                if failure.strip() == '': continue
                msg  = _("\tFailure").ljust(15)
                msg += failure
                payload += msg.rstrip()

        # output the rest of the messages
        for item in result.output_items:
            formatted = self.__output_one_item(item, options)
            if formatted:
                payload += '%s\n' % formatted

        return payload

    def __print_list_header(self):
        """ Prints a textual representation of header of sectool --list """
        print "%s\n%s %s %s\n%s" % ("-"*80,
                                   _("Name").ljust(self.list_w_name), 
                                   _("Description").ljust(self.list_w_desc), 
                                   _("Levels").ljust(self.list_w_levels),
                                   "-"*80)

    def __print_list_one_test(self, test_dsc, valid_levels):
        """ Prints one line of info about a test. """
        # if the description is too wide, append dots
        desc = test_dsc['HEADER']['DESCRIPTION'].ljust(self.list_w_desc)
        if '\n' in desc:
            desc = desc.replace('\n', ' ')

        if len(desc) > self.list_w_desc:
            desc = "%s..." % desc[:self.list_w_desc-3]

        # align corresponding levels
        str_levels = ""
        for i in xrange(1, ST_TestCase.max_level+1):
            if i in valid_levels:
                str_levels += "%d " % (i)
            else:
                str_levels += "  "

        # print as one string
        print "%s %s %s" % (test_dsc['HEADER']['NAME'].ljust(self.list_w_name)[:self.list_w_name],
                            desc,
                            str_levels)

    def output_test_list(self, tests):
        """ See superclass for docs """
        self.__print_list_header()
        for tst in tests:
            self.__print_list_one_test(tst.dsc, tst.valid_levels)

    def __print_info_one_test(self, test_dsc):
        """ Prints info about a test. """

        # FIXME - wrap description after 40 chars. Take a look at how yum does it.
        info_template = "%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n"
        print info_template % (_("Name").ljust(self.info_pad), test_dsc['HEADER']['NAME'], 
                                _("Script name").ljust(self.info_pad), test_dsc['HEADER']['FILE_NAME'],
                                _("Groups").ljust(self.info_pad), test_dsc['HEADER']['GROUPS'],
                                _("Levels").ljust(self.info_pad), test_dsc['HEADER']['LEVELS'],
                                _("Description").ljust(self.info_pad), test_dsc['HEADER']['DESCRIPTION'])

    def output_test_info(self, tests):
        """ See superclass for docs """
        for tst in tests:
            self.__print_info_one_test(tst.dsc) 

    def output_summary(self, total, passed, warn, error, fail, fatal):
        """ See superclass for docs """
        report_template = "%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n"

        print "\n"
        print "----- %s -----" % (_("Run report"))
        print report_template % (_("Total tests run").ljust(self.report_pad), total,
                                 _("Passed").ljust(self.report_pad), passed,
                                 _("Warnings").ljust(self.report_pad), warn,
                                 _("Errors").ljust(self.report_pad), error,
                                 _("Failures").ljust(self.report_pad), fail,
                                 _("Fatal failures").ljust(self.report_pad), fatal)

class NullFormatter(OutputFormatter):
    """ Null formatter. Eats all output. """

    def output_one_diff(self, test_name, diff, options, raw=True):
        return ""

    def output_one_result(self, results, options):
        return ""

    def output_test_list(self, tests):
        return ""

    def output_test_info(self, tests):
        return ""

    def output_summary(self, total, passed, warn, error, fail, fatal):
        return ""

    def result_callback(self, item, cb_data):
        return ""

    def test_result_header(self, test_name, raw=False):
        return ""

    def test_result_footer(self, test_name, test_result, raw=False):
        return ""

    def diff_result_footer(self, test_name, diff, raw=False):
        return ""

class HtmlFormatter(OutputFormatter):
    """ TBD """
    pass

