# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek

import exceptions

from i18n import _

class SectoolrcError(exceptions.Exception):
    """
    Raised on problems with loading the ~/.sectoolrc file
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("In file %s: %s (line: %s)" % (self.fname, self.msg, self.badline))

    def __init__(self, msg, fname, badline):
        """
        @param fname: name of the file this error occured in
        @type  fname: string 
        @param badline: the offending line
        @type  badline: string 
        """
        Exception.__init__(self)
        self.msg     = msg
        self.fname   = fname
        self.badline = badline

class MissingRequirementError(exceptions.Exception):
    """
    Raised on problems with loading the test such as missing dependencies
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Test %s tests package %s which is not installed, cannot run the test" % (self.test, ", ".join(dep for dep in self.deps)))

    def __init__(self, test, deps):
        """
        @param test: name of the test this error occured in
        @type  test: string 
        @param deps: a list of missings requirements
        @type  deps: list 
        """
        Exception.__init__(self)
        self.test = test
        self.deps = deps

class MissingDependencyError(exceptions.Exception):
    """
    Raised on problems with loading the test such as missing dependencies
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Test %s has missing dependencies: %s" % (self.test, ", ".join(dep for dep in self.deps)))

    def __init__(self, test, deps):
        """
        @param test: name of the test this error occured in
        @type  test: string 
        @param deps: a list of missings dependencies
        @type  deps: list 
        """
        Exception.__init__(self)
        self.test = test
        self.deps = deps

class ConfigValidationError(exceptions.Exception):
    """
    Raised on problems with validating contents of a config file
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Error in section %s of the config file %s: %s" % (self.section, self.file_name, self.message))

    def __init__(self, file_name, section, message):
        """
        @param file_name: full path to the config file
        @type  file_name: string 
        @param section: config file section in which the error occurred
        @type  section: string 
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.file_name = file_name
        self.section = section
        self.message = message

class ValidationError(exceptions.Exception):
    """
    Raised on problems with validating contents of a dsc file
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Validation error in section %s of file %s: %s" % (self.section, self.file_name, self.message))

    def __init__(self, file_name, section, message):
        """
        @param file_name: full path to the config file
        @type  file_name: string 
        @param section: config file section in which the error occurred
        @type  section: string 
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.file_name = file_name
        self.section = section
        self.message = message

class PeerError(exceptions.Exception):
    """
    Raised when there is a problem between a .dsc file and the worker script, such as:
    the worker script is missing
    the worker script is not executable
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Peer error for test %s: %s" % (self.basename, self.message))

    def __init__(self, basename, message):
        """
        @param basename: full path to the config file
        @type  basename: string 
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.basename = basename
        self.message = message

class ParseError(exceptions.Exception):
    """
    Raised when a script returns a line of output we cannot 
    parse.
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("Parse error: cannot parse \"%s\"" % (self.message))

    def __init__(self, message):
        """
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.message = message

class SecToolError(exceptions.Exception):
    """
    Generic error in SecTool
    """

    def __str__(self):
        """ The actual representation of the exception """
        return _("SecTool Error: %s" % (self.message))

    def __init__(self, message):
        """
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.message = message

class XmlError(exceptions.Exception):
    """
    Error that can happen while processing XML results file
    """
    
    E_TOO_MANY  = 0
    E_NOT_FOUND = 1

    def __str__(self):
        """ The actual representation of the exception """
        return _("XML results error %d: %s (processing %s)" % (self.err_id, self.message, self.path))

    def __init__(self, path, err_id, message):
        """
        @param message: explanation of the error
        @type  message: string 
        """
        Exception.__init__(self)
        self.path    = path
        self.message = message
        self.err_id  = err_id



