# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek

# secTool imports
import errors
from i18n import _

# stock python imports
import os
import sys
import re
import fnmatch
import bisect
import stat
import string
import exceptions
import select
import fcntl
import time
import rpm
import shutil

from UserDict import UserDict
from subprocess import (Popen, PIPE)
from ConfigParser import (SafeConfigParser, DuplicateSectionError, NoOptionError)
from unittest import (TestResult,
                      TestCase,
                      TestSuite)

print_stderr = True


def check_levels_validity(levels):
    for lvl in levels:
        if lvl not in range(ST_TestCase.min_level, ST_TestCase.max_level + 1):
            raise ValueError(_("Level specification allows only numbers in range %d to %d") %
                             (ST_TestCase.min_level, ST_TestCase.max_level))

    return True

class SectoolConfigParser(SafeConfigParser):
    """
    This is basically stock SafeConfigParser which is case sensitive. More changes might come later.
    """
    def optionxform(self, optionstr):
        """ preserves case sensitivity unlike superclass definition """
        return optionstr

class DescSection(object):
    """
    One section of the description file, such as [HEADER] or [DEFAULT]

    One can access option in section as class attributes, like:

    >>> section = DescSection("HEADER", some_parser)
    >>> section["DESCRIPTION"]

    All sections are also iterable:

    >>> section = DescSection("HEADER", some_parser)
    >>> for option in section: print option
    """

    # FIXME - convert to UserDict, too?
    def __init__ (self, name, parser):
        """
        @param name: name of the section
        @type name: string
        @param parser: a parser instance
        @type parser: ConfigParser
        """
        self.name = name
        self.parser = parser

    def __repr__(self):
        return "%s - %s" % (self.name, self.parser.items(self.name))

    def __getitem__ (self, name):
        """
        Returns value of the option referenced by the param name.

        @param name: name of the option
        @type name: string
        @rtype: string
        """
        item = self.parser.get(self.name, name)
        if item.startswith("\"") and item.endswith("\""):
            return item[1:-1]
        else:
            return item

    def __iter__(self):
        """
        Iterates over options in a section
        """
        return iter(self.parser.options(self.name))

class Description(UserDict):
    """
    The description file and access to its contents.

    One can accces an option from the description file like:

    >>> dsc["HEADER"]["DESCRIPTION"]

    One can also iterate over the sections:

    >>> for section is dsc: print section

    @todo: maybe we need a more generic validator
    """

    def __init__(self, fileName):
        """
        Constructs the instance based on a .dsc filename.

        @param fileName: a path to the .dsc file
        @type fileName: string
        @raises ValidationError: when the config file cannot be read
        @raises IOError: when the file referenced by the fileName parameter cannot be found
        """
        UserDict.__init__(self)

        if not os.path.exists(fileName):
            raise IOError(_("Test not found: %s" % fileName))

        self.fileName = fileName
        self.parser = SectoolConfigParser()
        try:
            self.parser.read(fileName)
        except exceptions.Exception, e:
            raise errors.ValidationError(self.fileName, "top level", e)

        # validate the description automagically
        self.__validate()

        # parse the groups tags into a set
        self.groups = set( self["HEADER"]["GROUPS"].split() )

    def __getitem__ (self, section):
        """
        Returns a DescSection instance of the section referenced by the param section.

        @param section: name of the section
        @type section: string
        @rtype: DescSection
        @raises KeyError: when there is no such section
        """
        if self.parser.has_section(section):
            return DescSection(section, self.parser)
        else:
            raise KeyError(_("No such section: %s" % (section)))

    def __repr__(self):
        """
        Textual representation in the form of: filename (test name) -> test script name
        @rtype: string
        """
        return "<%s (%s) -> %s>" % (self.fileName, self["HEADER"]["NAME"], self["HEADER"]["FILE_NAME"])

    def __iter__(self):
        """
        Iterates over sections in the dsc file
        """
        return iter(self.parser.sections())

    def __checkHeader(self):
        """
        Checks header if it includes all required options.
        True if includes, raises Validation Error if not

        @rtype: bool
        @raises ValidationError: if does not contain all the required options
        @todo: disallow other sections than the mandatory ones
        """
        required_set = set([ "NAME", "FILE_NAME", "DESCRIPTION", "LEVELS", "GROUPS" ])
        # fixme - check if we contain anything more than allowed directives, yell if so
        #optional_set = set([ "DEPS", "NEED_DATA_DIR" ])

        opt_set = set( [ opt for opt in self.parser.options('HEADER') ] ) 
        # check if all required are present
        missing_required = required_set - opt_set
        if len(missing_required) != 0:
            raise errors.ValidationError(self.fileName, "HEADER", _("Missing options %s" % (", ".join( [ opt for opt in missing_required ] ))))

        return True

    def __checkOptionValues(self):
        """
        If a certain option has a value of a certain type (like LEVEL - range of integers, etc.),
        a check should be done here

        @rtype: bool
        @raises ValidationError: if does not contain all the required options
        """

        # name should be one word, no whitespace (CLI friendly)
        if len([ ws for ws in string.whitespace if ws in self['HEADER']['NAME'] ]) != 0:
            raise errors.ValidationError(self.fileName, 'HEADER', _("Test name \"%s\" contains whitespace - this is ugly") % (self['HEADER']['NAME']))
        
        # if present, NEED_DATA_DIR should be either 0 or 1
        try:
            if self['HEADER']['NEED_DATA_DIR'] not in ['0', '1']:
                raise errors.ValidationError(self.fileName, 'HEADER', _("The NEED_DATA_DIR directive should be either 0 or 1, now it's %s") % (self['HEADER']['NEED_DATA_DIR'] ))
        except NoOptionError: pass # No directive, no checking needed

        # levels should contain only integers - this checking is done in __init__ along with converting them to integers

        return True

    def __validate(self):
        """
        True if validates, raises Validation Error if not

        @rtype: bool
        @raises ValidationError: if does not contain all the required sections
        """
        # If there's more like this, rewrite in a sane way
        if not self.parser.has_section("HEADER"):
            raise errors.ValidationError(self.fileName, "HEADER", _("No HEADER section"))

        self.__checkHeader()
        self.__checkOptionValues()

        return True

    def getLevelOptions(self, level):
        """
        Returns a list of all options plus all the defaults
        for a given numeric level

        @param level: the level we ask info about
        @type level: integer
        @rtype: list
        """
        return self["LEVEL_%d" % (level)]

    def has_level_section(self,  level):
        return self.parser.has_section("LEVEL_%d" % (level))

class OutputItem(object):
    """
    A structure-like abstraction of a message, its ID and severity
    """
    t_error     = 1
    "@ivar: the message severity is ERROR"
    t_warning   = 2
    "@ivar: the message severity is WARNING"
    t_hint      = 3
    "@ivar: the message severity is HINT"
    t_info      = 4
    "@ivar: the message severity is INFO"

    def __eq__(self, other):
        if self.type == other.type and self.msg == other.msg:
            return True
        return False

    def __str__(self):
        return "type: %s id: %s msg: %s" % (self.type, self.msg_id, self.msg)

    def __hash__(self):
        return hash(self.msg) ^ hash(self.msg_id) ^ hash(self.type)
    
    def __init__(self, msg_id, p_type, msg):
        """
        @param msg_id: a numeric ID of the message
        @type msg_id: integer
        @param p_type: message severity
        @type p_type: one of the class variables t_error, t_warning, ..
        @param msg: the actual message
        @type msg: string
        """
        self.msg_id   = msg_id
        self.type     = p_type
        self.msg      = msg

class ST_TestResultDiff(object):
    """
    Represents a diff between two ST_TestResult objects

    Added messages are being put ingo self.added, removed into
    self.removed. The self.are_same predicate tells whether the
    two results differ.
    """
    def __init__(self, before, after):
        """
        @param before: the "old" test result
        @type before: string
        @param after: the "new" test result
        @type after: string
        """
        self.before = before
        self.after  = after

        self.are_same = True
        self.added = set()
        self.removed = set()

    def __str__(self):
        return "Are same: %s Added: %d Removed: %d" % (self.are_same, len(self.added), len(self.removed))
        
    def diff(self):
        """ 
        Do the actual diff.
        """
        before_set = set([ item for item in self.before.output_items if item.type in (OutputItem.t_error, OutputItem.t_warning) ])
        after_set = set([ item for item in self.after.output_items if item.type in (OutputItem.t_error, OutputItem.t_warning) ])

        self.added = after_set - before_set
        self.removed = before_set - after_set

        self.are_same = (self.before.return_code == self.after.return_code) and \
                        (self.before.result == self.after.result) and \
                        (len(self.added) == 0 and len(self.removed) == 0)

        return self.are_same


class ST_TestResult(TestResult):
    """
    A container object that is able to parse messages from a test and store them
    in appropriate categories.
    
    See secTool documentation for explanation of the categories.
    The category "failure" indicates a bug in the test, the test is interrupted on failure.
    """
    NOT_YET_RAN = -1
    "@ivar: indicates that the test has no result so far"

    r_passed  = 0
    "@ivar: the worst result encountered was HINT or INFO"
    r_warning = 1
    "@ivar: the worst result encountered was WARNING"
    r_error   = 2
    "@ivar: the worst result encountered was ERROR"
    r_failure = 3
    "@ivar: the worst result encountered was FAIL"
    r_fatal   = 4
    "@ivar: cancels execution of the rest of the tests"

    def __init__(self):
        TestResult.__init__(self)
        self.return_code = ST_TestResult.NOT_YET_RAN
        self.pattern     = re.compile(("(?P<magic>[WEHI])(?P<id>\d+){0,1} : (?P<msg>.*)"))

        self.output_items = []  # 'expected' test output - warnings, errors etc.
        self.unmatched    = []  # output we could not fit anywhere - should be a bug in the test
        self.stderr       = []  # stderr of the test

        self.result       = self.r_passed

    def __repr__(self):
        """ 
        A textual representation suitable for parsing
        @rtype: string
        """
        return "Messages: %d, return code: %d, result: %s" % (len(self.output_items), self.return_code, self.result)

    def addOutputItem(self, item):
        """
        Add a filled-up OutputItem to self

        @param item: the item to add
        @type item: OutputItem
        """
        mapping = { OutputItem.t_error   : self.addError,
                    OutputItem.t_warning : self.addWarning,
                    OutputItem.t_hint    : self.addHint,
                    OutputItem.t_info    : self.addInfo }

        mapping[item.type](item.msg_id, item.msg)

    def __addMsg(self, msg_id, severity, msg):
        newitem = OutputItem(msg_id, severity, msg)
        self.output_items.append(newitem)
        return newitem

    def addWarning(self, msg_id, msg):
        """
        Add a warning to self and adjust self.result accordingly

        @param msg_id: numeric message ID
        @type msg_id: integer
        @param msg: the actual parsed message
        @type msg: string
        """
        if self.result < self.r_warning:
            self.result = self.r_warning

        return self.__addMsg(msg_id, OutputItem.t_warning, msg)


    def addHint(self, msg_id, msg):
        """
        Add a hint to self and adjust self.result accordingly 

        @param msg_id: numeric message ID
        @type msg_id: integer
        @param msg: the actual parsed message
        @type msg: string
        """
        return self.__addMsg(msg_id, OutputItem.t_hint, msg)

    def addError(self, msg_id, msg):
        """
        Add a error to self and adjust self.result accordingly 

        @param msg_id: numeric message ID
        @type msg_id: integer
        @param msg: the actual parsed message
        @type msg: string
        """
        if self.result < self.r_error:
            self.result = self.r_error

        return self.__addMsg(msg_id, OutputItem.t_error, msg)

    def addInfo(self, msg_id, msg):
        """
        Add a info message to self and adjust self.result accordingly 

        @param msg_id: numeric message ID
        @type msg_id: integer
        @param msg: the actual parsed message
        @type msg: string
        """
        return self.__addMsg(msg_id, OutputItem.t_info, msg)

    def addUnmatched(self, msg):
        """
        Add an other message to self

        @param msg: the actual parsed message
        @type msg: string
        """
        return self.unmatched.append(msg)

    def __guessMsgType(self, magic):
        """
        Returns correct add-to function for a given token.

        @param magic: one of W,H,I,E for various types of messages
        @type magic: string
        @rtype: function that should be used to add the message
        @raises KeyError: when the magic is not understood
        """

        mapping = { "W" : self.addWarning,
                    "H" : self.addHint,
                    "I" : self.addInfo,
                    "E" : self.addError }
        return mapping[magic]

    def process_one_line(self, line, ui_callback):
        """
        Processes one line at time and adds them to appropriate categories.

        @param line: line of output from test's stdout
        @type line: string
        @param ui_callback: function that belongs to a UI and processes messages from the test
        @type ui_callback: callable
        @raises ParseError: when it cannot parse a message correctly.
        """
        line = line.strip()
        if line == '': return # discard whitespace-only lines

        match = self.pattern.match(line)
        if match:
            parsed = match.groupdict()
            try:
                fn = self.__guessMsgType(parsed["magic"])
                new_item = fn(parsed["id"], parsed["msg"])
                ui_callback(new_item)
            except KeyError:
                raise errors.ParseError(line)
        else:
            self.addUnmatched(line)     # should we print these upon --debug ? Or fail ?

        
    def post_process(self, stdout, stderr, ui_callback):
        """
        Called after the test finishes to finish output and result processing

        @param stdout: lines of output from test's stdout
        @type stdout: list of strings
        @param stderr: lines of output from test's stderr
        @type stderr: list of strings
        """
        # the test is marked as FAILED when the return code != 0
        if self.return_code != 0:
            if self.return_code == 2:
                self.result = self.r_fatal
            else:
                self.result = self.r_failure

        # stderr is a punk area, so we just pass the reference
        self.stderr = stderr

        # stdout talks in 'protocol' -> parse the messages
        for line in stdout:
            self.process_one_line(line, ui_callback)

class ST_TestCase(TestCase):
    """
    One actual test.
    
    Can be instantiated by passing a Description object, i.e.:

    >>> dsc = Description("test.dsc")
    >>> tcase = ST_TestCase(dsc)

    For most of the information, see unittesting.TestCase

    The test can be executed by calling the instance itself:

    >>> tcase()

    or by calling the run() method:

    >>> tcase.run()
    """
    default_level   = 0
    min_level       = 1
    max_level       = 5
    data_dir_perms  = '0600'

    def __init__(self, dsc, environ={}):
        """
        @param dsc: a Description instance 
        @type dsc: string or Description instance
        @param environ: initial environment variables to pass to the test
        @type environ: dictionary
        """
        self.dsc = dsc

        # parse the valid levels string into a real list 
        # the name valid_levels is a bit misleading, we validate them now
        try:
            self.valid_levels = [ int(level) for level in self.dsc["HEADER"]["LEVELS"].split() ]
            check_levels_validity(self.valid_levels)
        except ValueError:
            raise errors.ValidationError(self.dsc.fileName, 'HEADER', _('Level specification contains a non-integer'))
        self.valid_levels.append(ST_TestCase.default_level)

        self.__generate_default_sections()
        self.setLevel(ST_TestCase.default_level)
        self.max_level = max(self.valid_levels)

        # every test needs a result...
        self.result = None

        # any custom environment changes go here
        self.environ = environ

        # a callback and its data
        self._callback = None
        self._cb_data  = None

        TestCase.__init__(self, methodName='runTest')

    def __generate_default_sections(self):
        # add sections with default-only values and the default runlevel
        for level in self.valid_levels:
            str_level = "LEVEL_%d" % (level)
            try:
                self.dsc.parser.add_section(str_level)
            except DuplicateSectionError: pass

    def assert_executable(self, test_base):
        """ see if the worker script exists and can be run """
        self.worker_path = os.path.join(test_base, self.dsc["HEADER"]["FILE_NAME"])
        if not os.access(self.worker_path, os.X_OK):
            raise errors.PeerError(self.dsc.fileName, _("The peer %s is either missing or not executable" % (self.worker_path)))

    def change_valid_levels(self, valid_levels):
        self.valid_levels = valid_levels
        self.__generate_default_sections()

    def __repr__(self):
        """ A textual representation """
        return str(self.dsc)

    def __str__(self):
        return self.__repr__()

    def __call__(self):
        """ Run the testsuite by calling the class """
        return self.run()

    def __lt__(self, other):
        """ Lexicographic comparing """
        return self.dsc.fileName < other.dsc.fileName

    def __gt__(self, other):
        """ Lexicographic comparing """
        return self.dsc.fileName > other.dsc.fileName

    def __which(self, program):
        """
        Returns True if program is in the search path, False if not

        @param program: the binary to look for
        @type program: string
        @rtype: boolean
        """
        if 'PATH' in self.environ:
            sys_paths = self.environ['PATH'].split(':')
        elif 'PATH' in os.environ:
            sys_paths = os.environ['PATH'].split(':')
        else:
            sys_paths = os.defpath.split(':')[1:]
        
        sys_paths.extend(["/sbin", "/usr/sbin"])
        for sysdir in sys_paths:
            if os.path.isfile(os.path.join(sysdir, program)):
                return True

        return False

    def checkForDeps(self, dsc):
        """
        Returns a list of dependencies not found is the given path

        @param dsc: - the Description instance to look at
        @type dsc: Description
        @rtype: list
        """
        try:
            dep_list = dsc["HEADER"]["DEPS"]
        except NoOptionError:
            return []  # no deps section -> no missing deps

        return [ program for program in dep_list.split() if self.__which(program) == False ]

    def check_requires(self, dsc):
        try:
            dep_list = dsc["HEADER"]["REQUIRES"]
        except NoOptionError:
            return []  # no deps section -> no missing deps

        ts = rpm.TransactionSet()
        return [ pkg for pkg in dep_list.split() if ts.dbMatch('name', pkg).count() == 0 ]


    def run(self):
        """
        runs the testcase with setUp and tearDown and stores 
        the test result int self.result

        @rtype: ST_TestResult
        """
        self.setUp()
        self.runTest()
        self.tearDown()

        return self.result

    def defaultTestResult(self):
        """ 
        Returns the default result storage, more or less implemented
        for compatibility with superclass 

        @rtype: ST_TestResult
        """
        return self.result

    def updateEnviron(self, **kwargs):
        """
        Allows to update environment independently of the .dsc script

        >>> updateEnviron(myprom="myvalue", myotherprom="foo")
        """
        self.environ.update(kwargs)

    def setUp(self):
        """
        Does preparation works before the test can be run - constructs environment, creates data dir
        if needed, etc.

        @raises IOError: If data directory was requested and cannot be created
        @raises IOError: If data directory exists from a previous run, but has bad permissions

        @todo: handle the IOErrors in the caller - don't run the test and set it to fail
        @todo: the code is ugly. Really ugly.
        """
        TestCase.setUp(self)

        # check for dependencies
        missing_deps = self.checkForDeps(self.dsc)
        if (len(missing_deps) != 0):
            raise errors.MissingDependencyError(self.dsc["HEADER"]["NAME"], missing_deps)

        # check for requirements
        missing_reqs = self.check_requires(self.dsc)
        if (len(missing_reqs) != 0):
            raise errors.MissingRequirementError(self.dsc["HEADER"]["NAME"], missing_reqs)

        # construct an enviroment from the header and level-specific values
        self.environ.update(dict([ (key, self.dsc["HEADER"][key]) for key in self.dsc["HEADER"] ]))
        self.environ.update(dict([ (key, self.dsc[self.level_str][key]) for key in self.dsc[self.level_str] ]))
        # tests need env. var. LANG to be localized
	self.environ.update( { "LANG" : os.getenv("LANG", "en_US.utf8")} )
        # finally, any specific tweaks passed on
        self.environ.update(dict([ (key, self.environ[key]) for key in self.environ.keys()] ))

        # if the test specifies a data dir, check if it's created, create if not
        if self.dsc.parser.has_option('HEADER','NEED_DATA_DIR') and self.dsc['HEADER']['NEED_DATA_DIR'] == '1':
            self._create_data_dir(os.path.join(self.environ['TDATA_DIR_BASE'], self.dsc['HEADER']['NAME']))

    def _create_data_dir(self, data_dir):
        if os.path.exists(data_dir):
            statres = os.lstat(data_dir)
            perms = oct(stat.S_IMODE(statres[stat.ST_MODE]) & 0777)
            if self.data_dir_perms != perms:
                raise IOError(_("Bad permissions on test data dir %s: got %s, expected %s" % (data_dir, str(perms), self.data_dir_perms)))
        else:
            try:
                os.makedirs(data_dir, int(self.data_dir_perms, 8))
            except:
                raise IOError(_("The data directory %s cannot be created (permissions=%s)" % (data_dir, self.data_dir_perms)))
        # pass the path to the test-specific data directory to the test
        self.environ.update( { 'TDATA_DIR' : data_dir } )

    def remove_data_dir(self):
        data_dir = os.path.join(self.environ['TDATA_DIR_BASE'], self.dsc['HEADER']['NAME'])
        try:
            shutil.rmtree(data_dir)
        except OSError: # No such file or directory - just skip
            pass

    def tearDown(self):
        """ Called after the test run to do clean-up """
        TestCase.tearDown(self)

    def setLevel(self, level):
        """
        Sets the runlevel for test. Level must be integer and valid
        for the test (wrt .dsc file).

        @param level: the level to set
        @type level: integer
        @raises ValueError: The level is not valid for this test
        @raises TypeError: The level was not given as an integer
        """
        try:
            int(level)
        except ValueError:
            raise TypeError(_("Level must be an integer"))
        if level not in self.valid_levels and not self.dsc.has_level_section(level):
            raise ValueError(_("Level %d is not valid for this test" % (level)))

        self.level = level
        self.level_str = "LEVEL_%d" % (self.level)

    def runTestLevel(self, level):
        """
        Runs the test directly on a certain level

        @param level: the level to set
        @type level: integer
        @raises ValueError: The level is not valid for this test
        @raises TypeError: The level was not given as an integer
        """
        self.setLevel(level)
        return self.run()

    def __set_nonblocking(self, fd):
        flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    def __flush_buffer(self, buffer):
        buf_content = ""
        buf_new = None

        while buf_new != "":
            buf_new = buffer.read()
            buf_content += buf_new

        return buf_content

    def runTest(self):
        """
        Runs the test worker itself and passes its stdout and stderr for processing.
        Does not set the environment or clean up, only calls the worker. 

        @rtype: ST_TestResult
        """
        # every test should have its own, new, shiny result
        self.result = ST_TestResult()

        try:
            test = Popen(self.worker_path,
                         stdout=PIPE,
                         stderr=PIPE,
                         close_fds=True,
                         env=self.environ)
        except OSError, e:
            print >>sys.stderr, _("Test execution failed"), e
        else:
            self.__set_nonblocking(test.stdout)
            self.__set_nonblocking(test.stderr)

            select_list = [ test.stderr, test.stdout ]
            t_stderr = ""
            t_stdout = ""
            newlines = []
            t_stderr_new = ""

            while test.poll() == None:
                ready = select.select(select_list, [], [], 0) 

                if test.stderr in ready[0]:
                    t_stderr_new += test.stderr.read()
                    if t_stderr_new != "":
                        if print_stderr == True and '\n' in t_stderr_new:
                            print >>sys.stderr, "stderr from %s: %s" % (self.dsc['HEADER']['NAME'], t_stderr_new)
                            t_stderr += t_stderr_new
                            t_stderr_new = ""

                if test.stdout in ready[0]:
                    t_stdout += test.stdout.read()
                    newlines = t_stdout.split('\n')

                    if t_stdout[-1] != '\n':
                        t_stdout = newlines[-1] 
                        newlines = newlines[:-1]
                    else:
                        t_stdout = ""

                    for nl in newlines:
                        self.result.process_one_line(nl, self.callback)

                time.sleep(0.3) 

            t_stdout += self.__flush_buffer(test.stdout)
            t_stderr += self.__flush_buffer(test.stderr)

            self.result.return_code = test.returncode
            self.result.post_process(t_stdout.split('\n'),
                                     t_stderr.split('\n'),
                                     self.callback)

        return self.result

    def refreshTest(self):
        """
        Removes any data owned by a test and runs it just to recreate them
        """

        self.remove_data_dir()
        self.run()

    def _exc_info(self):
        """
        Standard unittest prints out execution info in a form of traceback.
        We print a brief info about the test being run
        @rtype: string
        """
        return self.__repr__()

    def register_callback(self, callback, cb_data):
        if not callable(callback):
            raise TypeError(_("The callback must be callable"))
        self._callback = callback
        self._cb_data  = cb_data  

    def callback(self, item):
        if self._callback:
            return self._callback(item, self._cb_data)

        return None

 

class ST_TestSuite(TestSuite):
    """
    A collection of tests either returned by ST_TestLoader or aggregated manually.

    Most of the hard work is done in the superclass - unittest.TestSuite
    """
    def __init__(self):
        TestSuite.__init__(self)

    def __add__(self, other):
        """
        Add tests from an other collection to self

        @param other: the suite to add from
        @type other: ST_TestSuite
        """
        for test in other:
            self.addTest(test)

    def __len__(self):
        """
        Returns the number of tests in the testsuite

        @rtype: integer
        """
        return len(self._tests)

    def __call__(self):
        """
        Runs all the test in the suite by calling it

        >>> st = ST_TestSuite()
        >>> st.add_tests(test_list)
        >>> st()
        """
        self.run()

    def __str__(self):
        return "Total %d tests: %s" % (self.countTestCases(), '\n'.join([ str(t) for t in self._tests]))

    def run(self, level=ST_TestCase.default_level):
        """
        Runs all the test in suite, collecting results. Stops on
        first fatal error. Returns a list of results
        
        @param level: level on which the suite runs
        @type level: integer
        @rtype: list
        """
        return [ test.runTestLevel(level) for test in self._tests ]

    def addTest(self, test):
        """
        Adds one test to the collection

        @param test: the test to add
        @type test: ST_TestCase
        @raises TypeError: if the test param is not of ST_TestCase type
        """
        if not callable(test):
            raise TypeError(_("The test to add must be callable"))
        if not isinstance(test, ST_TestCase):
            raise TypeError(_("Test to add must be of type ST_TestCase"))

        bisect.insort(self._tests, test)

    def filter(self, filter_func, **kwargs):
        """
        Return a new TestSuite containing tests for which the filter_func returns True. All
        others are discarded. All keyword-arguments are passed to the function.
        
        @param filter_func: the filtering function
        @type filter_func: function
        @rtype: ST_TestSuite
        @raises TypeError: If the filter function is not callable
        """
        if not callable(filter_func):
            raise TypeError(_("The filter function must be callable"))

        suite = ST_TestSuite()
        [ suite.addTest(test) for test in self._tests if filter_func(test.dsc, **kwargs) ]
        return suite

    def filterInPlace(self, filter_func, **kwargs):
        """
        Filters tests in this testSuite accordinf to the filter_func. Only those
        for which the filter_func returns true are left in the testSuite
        
        @param filter_func: the filtering function
        @type filter_func: function
        @raises TypeError: If the filter function is not callable
        """
        if not callable(filter_func):
            raise TypeError(_("The filter function must be callable"))
        self._tests = [ test for test in self._tests if filter_func(test, **kwargs) ]

    def updateEnviron(self, environ=None, **kwargs):
        """
        Updates environment of all tests in the suite
        
        @param environ: a dictionary to update from
        @type environ: dictionary
        """
        env = dict(**kwargs)
        env.update(environ)
        
        [ test.environ.update(env) for test in self._tests ]

    def register_callback(self, callback, cb_data):
        [ test.register_callback(callback, cb_data) for test in self._tests ]
        

class ST_TestLoader(object):
    """
    Finds tests (and returns ST_TestSuite) or finds test descriptions
    (and returns list of) in a given path.

    The tests can be selected by a filter method which is given a
    Description instance as a parameter.

    Idea based on unittest.TestLoader

    Also provides a set of common filters - by name, group, ...see filter* methods
    """

    dsc_suffix = "*.dsc"
    "@ivar the standard glob pattern to find description files"

    def __init__(self, filter_func=lambda x:True, level_config = None):
        """
        @param filter_func: When using the filer_func, it must return 
        true for test wa want to keep and False for test we want to drop.
        @type filter_func: function
        """

        self.level_override = dict()
        if level_config:
            self.level_override = self.__get_custom_levels(level_config)

        self.setFilter(filter_func)

    def __get_custom_levels(self, level_config):
        try:
            f = open(level_config)
        except IOError:
            return dict()       # no big deal

        level_override = dict()
        for line in f.readlines():
            splitted = line.split()
            try:
                levels = [ int(level) for level in splitted[1:] ]
                check_levels_validity(levels)
                levels.append(ST_TestCase.default_level)
                levels.sort()
            except ValueError:
                raise errors.SectoolrcError("level specification contains bad specification of level", level_config, line.strip())
            if not 1 <= len(levels) <= 6:
                raise errors.SectoolrcError("level specification contains bad number of levels", level_config, line.strip()) 

            level_override[splitted[0]] = levels

        return level_override

    def setFilter(self, filter_func):
        """
        Sets the filter function to a new one. filter_func must be callable

        @param filter_func: the filtering function
        @type filter_func: function
        @raises TypeError: when the filter_func is not callable
        """
        if not callable(filter_func):
            raise TypeError(_("The filter function must be callable"))
        else:
            self.filter = filter_func

    @staticmethod
    def filterIncludeTags(dsc, tags):
        """
        Filter only testcases that include one of the tags from the list

        @param dsc: - the Description instance to look at
        @type dsc: Description
        @param tags: - if description contains one of these tags, returns True
        @param tags: list
        @rtype boolean
        """
        return (dsc.groups & set(tags)) != set([])

    @staticmethod
    def filterNameContains(dsc, name):
        """
        Filter only testcases whose name contain the parameter name

        @param dsc: - the Description instance to look at
        @type dsc: Description
        @param name: - if description's name contains this string, return true
        @param name: string
        @rtype boolean
        """
        return name in dsc["HEADER"]["NAME"]

    @staticmethod
    def filterNameEquals(dsc, name):
        """
        Filter only testcases whose name equals the parameter name

        @param dsc: - the Description instance to look at
        @type dsc: Description
        @param name: - if description's name equals this string, return true
        @param name: string
        @rtype boolean
        """
        return name == dsc["HEADER"]["NAME"]

    @staticmethod
    def filterDescriptionContains(dsc, dsc_str):
        """
        Filter only testcases whose description contains the string dsc

        @param dsc: - the Description instance to look at
        @type dsc: Description
        @param dsc_str: - if description contains this string, return true
        @param dsc_str: string
        @rtype boolean
        """
        return dsc_str in dsc["HEADER"]["DESCRIPTION"]

    def findDescriptions(self, path):
        """
        Finds description files in a given path and returns a list of

        @param path: the path to search at
        @type path: string
        @rtype: list
        """
        found = list()
        
        if not os.path.exists(path): 
            raise IOError("Cannot find path %s (cwd=%s)" % (path, os.getcwd()))

        for dirpath, dirs, files in os.walk(path):
            found.extend( [ os.path.join(dirpath, f) for f in files if fnmatch.fnmatch(f, ST_TestLoader.dsc_suffix) ] )

        return found

    def loadFromList(self, load_list, test_path):
        """
        Constructs a ST_TestSuite from a list of description files

        @param load_list: a list of dsc files
        @name load_list: list
        @rtype: ST_TestSuite
        """
        suite = ST_TestSuite()

        for fullpath in load_list:
            dsc = self.loadTest(fullpath)
            if dsc != None:
                tcase = ST_TestCase(dsc)
                tcase.assert_executable(test_path)
                if dsc['HEADER']['NAME'] in self.level_override:
                    tcase.change_valid_levels(self.level_override[dsc['HEADER']['NAME']])
                suite.addTest(tcase)

        return suite

    def loadFromPath(self, dsc_path, test_path):
        """
        Finds tests in a given path, returns a ST_TestSuite instance

        @param path: the path to search at
        @type path: string
        @rtype: ST_TestSuite
        """
        return self.loadFromList(self.findDescriptions(dsc_path), test_path)

    def loadTest(self, fullpath):
        """
        Loads one test based on path to description file and
        filters it through the filter function provided at __init__ time

        @param fullpath: the path to the .dsc file
        @type fullpath: string
        @rtype: Description
        """
        dsc = Description(fullpath)
        if self.filter(dsc) == False:
            return None

        return dsc

class SecToolConf(UserDict):
    """
    A UserDict extension that is used to parse the config file
    (default at /etc/sectool/sectool.conf)
    """

    required = [ "PATHS", "RESULTS", "MAIL" ]
    userfile = ".sectoolrc"

    def __init__(self, fileName):
        """ Open and parse the config file """
        UserDict.__init__(self)
        self.fileName = fileName
        self.parser = SafeConfigParser()
        if len(self.parser.read(fileName)) == 0:
            raise IOError("Could not open the config file %s" % (fileName))

    def __getitem__ (self, section):
        if self.parser.has_section(section):
            return DescSection(section, self.parser)
        else:
            raise KeyError(_("No such section: %s" % (section)))

    def __iter__(self):
        return iter(self.parser.sections())

    def save_custom_levels( self, level_override ):
        
        file = os.path.join(os.path.expanduser("~"), self.userfile )
        try:
            f = open( file, 'w' )
        except IOError:
            raise IOError( "Could not create the config file %s" % ( file ) )

        for test in level_override:
            f.write( test + " " + ' '.join( [`level` for level in level_override[ test ] ] ) + '\n' )
        
        f.close()

    def validate(self):
        """
        True if validates, raises Validation Error if not
        FIXME - Need concrete specs on when is a description file valid
        """
        # If there's more like this, rewrite in a sane way
        for section in SecToolConf.required:
            if not self.parser.has_section(section):
                raise errors.ConfigValidationError(self.fileName, section, _("No %s section" % (section)))

        return True

