# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek

# FIXME - should be validated against DTD
# FIXME - could speed up by adding 'total' or 'warnings', 'errors' etc. attribute to <messages>
# FIXME - store language in the xml document itself, don't compare different languages
# FIXME - store sectool version-release in the xml document itself

import os
import errno
from xml.dom import minidom as minidom
from xml.parsers.expat import ExpatError as ExpatError

import scheduler
from errors import XmlError as XmlError

ENCODING = 'utf-8'  

class XmlOutput(object):
    def __init__(self, path):
        self.path = path
        try:
            self.parsed_xml = minidom.parse(path)
        except ExpatError, e:
            if e.lineno == 1 and e.offset == 0 and e.code == 3:
                self._init_empty_file()
            else:
                raise e
        except IOError, e:
            if e.errno == errno.ENOENT:
                self._init_empty_file()
            else:
                raise e
        else:
            self.results_root = self.parsed_xml.getElementsByTagName('results')[0]
            self._update_test_list()

    def _init_empty_file(self):
        self.parsed_xml = minidom.Document()
        self.results_root = self.parsed_xml.createElement('results')
        self.parsed_xml.appendChild(self.results_root)
        self.test_list  = []

        self._output_pretty_xml()

    def _update_test_list(self):
        # FIXME - this might be doable with std. python operations, which might be faster
        #         as in, when appending new result, append the node to the end of the list etc.
        self.test_list = [ node for node in self.results_root.childNodes if node.nodeType == minidom.Element.ELEMENT_NODE ]

    def _decode_item(self, msg_node):
        # item id can be None for cases of Info messages
        try:
            item_id = int(msg_node.getAttribute('id'))
        except ValueError:
            item_id = None

        item = scheduler.OutputItem(item_id,
                                    int(msg_node.getAttribute('type')),
                                    msg_node.firstChild.nodeValue.strip())

        return item

    def __xml2result(self, xml_result):
        result = scheduler.ST_TestResult()
        result.result = int(xml_result.getElementsByTagName('result')[0].firstChild.nodeValue)
        
        msgs_node = xml_result.getElementsByTagName('messages')[0]
        for msg in [ node for node in msgs_node.childNodes if node.nodeType == minidom.Element.ELEMENT_NODE ]:
            result.addOutputItem(self._decode_item(msg))

        return result

    def __result2xml(self, test_name, test_result):
        doc = minidom.Document()
        test = doc.createElement('test')

        name = doc.createElement('name')
        name.appendChild(doc.createTextNode(test_name))
        test.appendChild(name)

        name = doc.createElement('return_code')
        name.appendChild(doc.createTextNode(str(test_result.return_code)))
        test.appendChild(name)

        result = doc.createElement('result')
        result.appendChild(doc.createTextNode(str(test_result.result)))
        test.appendChild(result)

        messages = doc.createElement('messages') 
        for item in test_result.output_items:
            item_elem = doc.createElement('msg')
            item_elem.setAttribute('type', str(item.type))
            item_elem.setAttribute('id', str(item.msg_id))
            item_elem.appendChild(doc.createTextNode(item.msg.decode(ENCODING, 'replace')))
            messages.appendChild(item_elem)

        test.appendChild(messages)

        return test

    def _diff_xml_results(self, new, old):
        result_diff = scheduler.ST_TestResultDiff(self.__xml2result(old), self.__xml2result(new))
        result_diff.diff()

        return result_diff

    def _output_pretty_xml(self):
        # FIXME - HACK!!!! 
        content = [ "%s\n" % (line) for line in self.parsed_xml.toprettyxml(encoding=ENCODING).split('\n') if line.strip() != '' ]
        open(self.path, 'w').writelines(content)
        os.chmod(self.path, 0600)

    def _name_matches(self, test_node, name):
        assert(len(test_node.getElementsByTagName('name')) == 1)     
        test_name = test_node.getElementsByTagName('name')[0]
        assert(test_name.firstChild.nodeType == minidom.Element.TEXT_NODE) 

        return test_name.firstChild.nodeValue.strip() == name

    def _substitute_test_result(self, test_name, xml_result):
        matches = [ test for test in self.test_list if self._name_matches(test, test_name) ] 
        if len(matches) > 1:
            raise XmlError(self.path, XmlError.E_TOO_MANY, 'Found %d tests with matching name %s, expected 1' % (len(matches), test_name))
        if len(matches) == 0:
            raise XmlError(self.path, XmlError.E_NOT_FOUND, 'No matching result found')
        match = matches[0]

        self.results_root.replaceChild(xml_result, match)        
        return match

    def write_result(self, test_name, test_result):
        result_diff = None

        xml_result = self.__result2xml(test_name, test_result)
        try:
            try:
                old_result = self._substitute_test_result(test_name, xml_result)
            except XmlError, e:
                if e.err_id == XmlError.E_NOT_FOUND:
                    self.results_root.appendChild(xml_result)
                else:
                    raise e
            else:
                result_diff = self._diff_xml_results(xml_result, old_result)
        finally:
            self._update_test_list()
            self._output_pretty_xml()

            return result_diff

