#!/usr/bin/python
# -*- coding: UTF8 -*-

## Copyright (C) 2007 Red Hat, Inc.
## Written by Jakub Hrozek <jhrozek@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


from scheduler import scheduler, errors
from scheduler.i18n import _
import actions
import output
import tuierrors

import sys
import copy
import traceback
import exceptions
from optparse import OptionParser, OptionValueError

APP_VERSION = "0.8.6"

def print_exc_plus():
    """
    Print the usual traceback information, followed by a listing of
    all the local variables in each frame.

    Code taken from the "Python Cookbook, 2nd edition, O'Reilly 2005"
    ISBN 0-596-00797-3
    """
    tb = sys.exc_info( )[2]

    while tb.tb_next:
        tb = tb.tb_next
    stack = [  ]
    f = tb.tb_frame

    while f:
        stack.append(f)
        f = f.f_back

    stack.reverse( )
    traceback.print_exc( )

    print "Locals by frame, innermost last"
    for frame in stack:
        print
        print "Frame %s in %s at line %s" % (frame.f_code.co_name,
                                            frame.f_code.co_filename,
                                            frame.f_lineno)
    for key, value in frame.f_locals.items( ):
        print "\t%20s = " % key,
        # we must _absolutely_ avoid propagating exceptions, and str(value)
        # COULD cause any exception, so we MUST catch any...:
        try:
            print value
        except:
            print "Could not print value"

class MyOptionParser(OptionParser):
    """
    Optparse does it own encoding even though you never ask it to do that. This 
    is insane and does not allow you to use UTF8 strings directly.
    """
    def print_help(self, file=None):
        if file == None:
            file=sys.stdout
        file.write(self.format_help())


class TuiOptions(object):
    """ A wrapper class for holding options about the text UI """

    def __init__(self):
        self.options  = None         # this holds the actual options from the user
        self.args = []               # this holds any additional arguments from the user

        #self.parser = OptionParser(prog="sectool",
        self.parser = MyOptionParser(prog="sectool",
                                   usage="Usage: %prog command [arguments]",
                                   version="%prog "+APP_VERSION)

        # options that tune sectool
        self.parser.add_option("-c", "--config", dest="config", type="string",
                action="store", metavar="CONFIG", default="/etc/sectool/sectool.conf",
                help=_("Specify a configuration file"))

        self.parser.add_option("-d", "--debug", dest="debug",
                action="store_true", default=False,
                help=_("Debug mode"))

        self.parser.add_option("", "--print-traceback", dest="print_traceback",
                action="store_true", default=False,
                help=_("""Print traceback for development purposes on encountering
an error that is otherwise handled gracefully"""))

        self.parser.add_option("-n", "--hint", dest="hint",
                action="store_true", default=False,
                help=_("Display hints when available"))

        self.parser.add_option("-m", "--mail", dest="mail", type="string",
                action="store", metavar="ADDRESS", default=False,
                help=_("Send email report"))

        self.parser.add_option("-g", "--groups", dest="group",
                action="callback", callback=self.varargs_callback,
                default=[],
                help=_("Filter tests by group"))

        self.parser.add_option("-f", "--diff", dest="diff",
                action="store_true", default=False,
                help=_("Display only diff against previous run (if available)"))

        self.parser.add_option("", "--include", dest="include",
                action="callback", callback=self.varargs_callback,
                default=[], metavar="INCLUDE",
                help=_("If running a level, include these tests that may be not defined in that level."))

        self.parser.add_option("", "--exclude", dest="exclude",
                action="callback", callback=self.varargs_callback,
                default=[], metavar="EXCLUDE",
                help=_("If running a level, exclude these tests that may be not defined in that level."))

        self.parser.add_option("", "--report", dest="report",
                action="store_true", default=False,
                help=_("Print a report about the run after it finishes"))

        self.parser.set_defaults(action_set=set())
        self.parser.set_defaults(level=scheduler.ST_TestCase.default_level)

        # actual sectool actions
        self.parser.add_option("-l", "--list",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("List tests and their level configuration"))

        self.parser.add_option("-i", "--info",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("Query info about a test"))

        self.parser.add_option("-L", "--level",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("""Run all tests in a collection at a certain level.
The level can be specified either as an integer (1-5) or string nickname
(Naive, Desktop, Network, Server, Paranoid). See sectool(8) for more details."""))

        self.parser.add_option("-r", "--run",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("Run selected tests"))

        self.parser.add_option("-a", "--auto", 
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("Perform the action defined in the config file"))

        self.parser.add_option('', "--clean",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("Cleans the XML cache which stores previous test results"))

        self.parser.add_option('', "--refresh-test",
                action="callback", callback=self.action_callback,
                dest="action_set",
                help=_("Removes any data owned by a test and runs it just to recreate them"))

    @staticmethod
    def action_callback(option, opt_str, value, parser):
        try:
            long_opt = str(option).split('/')[1]
        except IndexError:
            long_opt = opt_str  # long option only

        mapping = { "--list"  : actions.ST_ListTests,
                    "--info"  : actions.ST_InfoTests,
                    "--level" : actions.ST_RunLevel,
                    "--auto"  : actions.ST_Auto,
                    "--run"   : actions.ST_RunTests,
                    "--clean" : actions.ST_CleanCache,
                    "--refresh-test" : actions.ST_RefreshTest }

        string_level_map = { 'NAIVE'   : 1,
                             'DESKTOP' : 2,
                             'NETWORK' : 3,
                             'SERVER'  : 4,
                             'PARANOID': 5 }

        if not long_opt in mapping.keys():
            raise OptionValueError(_("Unknown action %s") % (long_opt))
        parser.values.action_set.add(mapping[long_opt])

        # for --level, we also want to store the next (required) argument specifying the actual level
        if long_opt == "--level":
            # a positive integer must follow
            try:
                parser.values.level = int(parser.rargs[0])
                if parser.values.level not in range(scheduler.ST_TestCase.min_level, scheduler.ST_TestCase.max_level + 1):
                    raise ValueError

                del parser.rargs[0]
            except ValueError:
                try:
                    parser.values.level = int(string_level_map[parser.rargs[0].upper()])
                except KeyError:
                    raise OptionValueError(_("Bad level specification - use --level <num> where <num> is between 1 and 5"))
            except IndexError:
                raise OptionValueError(_("Bad level specification - use --level <num> where <num> is between 1 and 5"))

    @staticmethod
    def varargs_callback(option, opt_str, value, parser):
        assert value is None
        done = 0
        value = []
        rargs = parser.rargs
        while rargs:
            arg = rargs[0]

            if ((arg[:2] == "--" and len(arg) > 2) or
                (arg[:1] == "-" and len(arg) > 1)):
                break
            else:
                value.append(arg)
            del rargs[0]

        setattr(parser.values, option.dest, value)

    def __str__(self):
        return "options: %s args: %s" % (self.options, self.args)

    def loadSchedulerOptions(self):
        " load config options from the scheduler "
        try:
            self.scheduler_opts = scheduler.SecToolConf(self.options.config)
            self.scheduler_opts.validate()
        except Exception, e:
            if self.options.debug:
                print_exc_plus()

            print >>sys.stderr, str(e)
            sys.exit(1)

    def parse_params(self, args):
        # do the parsing
        self.options, self.args = self.parser.parse_args()
        # possible error-checking would also go here
        return self.options, self.args 

    def get_action(self):
        """ Based on options provided, returns a constructor of a descendant of ST_Action """
        # user must select an action to run...
        if len(self.options.action_set) == 0:
            self.parser.print_help()
            self.parser.error(_("Must select at least one action"))
            
        # ...but not more than one...
        if len(self.options.action_set) > 1:
            # ...unless its --run --level (see manpage)
            if (actions.ST_RunLevel in self.options.action_set and
               actions.ST_RunTests in self.options.action_set and
               len(self.options.action_set) == 2):
                return actions.ST_RunTests
            else:
                self.parser.print_help()
                self.parser.error(_("Only one action can be selected at a time"))
        else: return copy.copy(self.options.action_set).pop() # yes, a hack



class SecToolTui(object):
    """ The base class of sectool text interface """

    def __bail_out(self, error, print_traceback = False):
        if print_traceback == True:
            print_exc_plus()

        print >>sys.stderr, error
        sys.exit(1)

    def run(self, options):
        action = None
        self.options = options

        # no other formatter as of now..
        fmt = output.StdoutFormatter()

        scheduler.print_stderr = self.options.options.debug

        # get an action based on the options and do it
        constructor = options.get_action()
        try:
            action = constructor(options, fmt)
            if action == None:
                raise tuierrors.TuiError(_("Could not instantiate action based upon %s") % (str(self.options)))
            action.doAction()
        except tuierrors.SecToolActionError, e:
            self.__bail_out(e, self.options.options.print_traceback)

        except tuierrors.PreReqError, e:
            self.__bail_out(e, self.options.options.print_traceback)

        except exceptions.Exception, e:
            if self.options.options.debug == True:
                print_exc_plus()
            else:
                traceback.print_exc( )
            


def main(args):
    # the base text UI
    base = SecToolTui()

    # options container, parse params
    options = TuiOptions()
    options.parse_params(args)
    options.loadSchedulerOptions()

    # do the work
    base.run(options)

if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt, e:
        print >> sys.stderr, _("Exiting on user request.\n")
        sys.exit(1)

