#!/usr/bin/env bash
. "${TESTS_DIRS}/bash_defs.sh"
# Copyright (C) 2008 Red Hat, Inc.
# Written by Dan Kopecek <dkopecek@redhat.com>

const E_NOFILE=13
const E_NORECORD=14
const E_NOLIST=15
const E_DIFF=16
const E_NEWLIST=17

# --- Error IDs ------- #
const ID_CHKLIST_FILECHANGED=1
const ID_CHKLIST_FILENOTFOUND=2
const ID_CHKLIST_NORECORD=3
const ID_CHKLIST_NOTFOUND=4
const ID_CKHLIST_NEW=5
const ID_KEYNOTENCRYPTED=6
const ID_OVPN_CONFDIRPERM=7
const ID_OVPN_WRONGPGRP=8
const ID_OVPN_WRONGPUSER=9
const ID_OVPN_PRIVKEYPERM=10

function check_env () {
    PROC="`mount | sed -n 's/.*on\s\(.*\)\stype proc.*/\1/p'`"
    if [[ -d "${PROC}" ]]; then
	return ${E_OK}
    else
	if `mount -t proc &> /dev/null`; then
	    if check_env; then
		return ${E_OK}
	    else
		return ${E_FAIL}
	    fi
	fi
    fi
    return ${E_FAIL}
}

function proc_getopt () { # 1 - cmdline, 2 - opt
    cat ${1} | tr '\0' '\t' | sed -n "s/.*${2}\t\([^\t]*\)\t.*/\1/p" && echo
}

function proc_getenv () {
    cat ${1} | tr '\0' '\n' | sed -n "s/^${2}=\(.*\)$/\1/p"
}

# <proc_statuc>
function proc_getuid () {
    sed -n 's/^Uid:\s\([[:digit:]]*\).*/\1/p' "${1}"
}

# <proc_status>
function proc_getgid () {
    sed -n 's/^Gid:\s\([[:digit:]]*\).*/\1/p' "${1}"
}

# <file> <list>
function sha256cmp () {
    if [[ ! -f "${1}" ]]; then
	return ${E_NOFILE}
    else
	local NEW="`sha256file \"${1}\"`"
    fi

    if [[ ! -f "${2}" ]]; then
	echo ${NEW}
	touch "${2}" && return ${E_NEWLIST}
	return ${E_NOLIST}
    fi

    local OLD=$(grep "SHA256(${1})" "${2}" 2> /dev/null)

    if [[ -n "${OLD}" ]]; then
	if [[ "${NEW}" == "${OLD}" ]]; then
	    return ${E_OK}
	else
	    echo ${NEW}
	    return ${E_DIFF}
	fi
    else
	echo ${NEW}
	return ${E_NORECORD}
    fi
}

# <file> <list> <save_flag> [<id>]
function check_sha256 () {
    local NEW
 
    NEW="`sha256cmp \"${1}\" \"${2}\"`"

    case "$?" in
	"${E_NEWLIST}")
	    echo ${NEW} > "${2}"
	    report 'WARNING' $ID_CKHLIST_NEW      "Created new SHA256 list." 
	    ;;
	"${E_NOLIST}")
	    report 'WARNING' $ID_CHKLIST_NOTFOUND "SHA256 list not found."
	    ;;
	"${E_OK}")
	    return ${E_OK}
	    ;;
	"${E_DIFF}")
	    report 'WARNING' $ID_CHKLIST_FILECHANGED "File \"${1}\" changed!"
	    if (( ${3} == 1 )); then
		grep -v "SHA256(${1})" "${2}" > "${2}.new"
		echo ${NEW} >> "${2}.new"
		mv "${2}.new" "${2}"
	    fi
	    return ${E_DIFF}
	    ;;
	"${E_NORECORD}")
	    report 'WARNING' $ID_CHKLIST_NORECORD "No SHA256 record for file \"${1}\""
	    if (( ${3} == 1 )); then
		echo ${NEW} >> "${2}"
	    fi
	    return ${E_NORECORD}
	    ;;
	"${E_NOFILE}")
	    report 'WARNING' $ID_CHKLIST_FILENOTFOUND "File \"${1}\" not found."
	    return ${E_NOFILE}
	    ;;
	*)
	    # FIXME: ...
	    return ${E_FAIL}
	    ;;
    esac
}

# __ovpn_getconf sshd_config HostKey
function __ovpn_getconf () {
    sed -n -e "s/^[^#]*[\s]*${2}\s\(.*\)/\1/p" ${1}
}

# __ovpn_getoptconf <cmdline> HostKey
function __ovpn_getoptconf {
    case "${2}" in
	-*)
	    proc_getopt "${1}" "${2}"
	    ;;
	*)
	    proc_getopt "${1}" "--${OPT}"
    esac
}

function __check_ovpn_config () {
    local PROCDIR=${1}
    local CONFFILE=${2}
    local CONFDIR="$(dirname ${CONFFILE})"
    local CLIENT=0

    # check permissions dir + config
    check_file_perm "${CONFDIR}"  ${OVPN_CONFIGDIR_PERM} ${OVPN_CONFIGDIR_OWNER} 1 $ID_OVPN_CONFDIRPERM "OpenVPN configuration directory"
    check_file_perm "${CONFFILE}" ${OVPN_CONFIG_PERM}    ${OVPN_CONFIG_OWNER}    1 $ID_OVPN_CONFDIRPERM "OpenVPN configuration file"

    if [[ -n "$(__ovpn_getconf "${CONFFILE}" client)" || -n "$(__ovpn_getoptconf "${PROCDIR}/cmdline" client)" || \
	  -n "$(__ovpn_getconf "${CONFFILE}" tls-client)" || -n "$(__ovpn_getoptconf "${PROCDIR}/cmdline" tls-client)" ]]; then
	CLIENT=1
    fi

    # resolve keyfile names + permission check
    # keyfiles from config

    for keyfile in `(__ovpn_getconf "${CONFFILE}" key; __ovpn_getoptconf "${PROCDIR}/cmdline" key) | sort | uniq`; do
	case "${keyfile}" in
	    /*)
		keyfile=$(rel2abs ${keyfile})
		;;
	    *)
	        CD="$(__ovpn_getconf "${CONFFILE}" cd)"
		if [[ -z "${CD}" ]]; then
		    CD="$(__ovpn_getoptconf "${PROCDIR}/cmdline" cd)"
		fi
	        keyfile=$(rel2abs "${CD}/${keyfile}")
		;;
	esac

	check_file_perm "${keyfile}"     ${OVPN_PRIVKEY_PERM} ${OVPN_PRIVKEY_OWNER}  1 $ID_OVPN_PRIVKEYPERM "OpenVPN private key"
	check_sha256    "${keyfile}"     "${TDATA_DIR}/${OVPN_SHA256_LIST}" "${OVPN_SAVE_SHA256_CHANGES}"

	if [[ -z "$(grep ENCRYPTED "${keyfile}")" && ${CLIENT} == 1 ]]; then
	    report 'WARNING' $ID_KEYNOTENCRYPTED "Your OpenVPN private key \"${keyfile}\" IS NOT encrypted."
	fi
    done

    # TODO: options check
    return ${E_OK}
}

function CHK_openvpn () {
    local PROCDIR=${2}
    local PKGNAME=${1}
    local CURRENT_CONF=$(proc_getopt "${PROCDIR}/cmdline" --config)
    local DEFAULT_CONF="${OVPN_DEFAULT_DIR}/openvpn.conf"
    local PUSER=$(uid2user `proc_getuid ${PROCDIR}/status`)
    local PGROUP=$(gid2user `proc_getgid ${PROCDIR}/status`)

    if [[ ${PUSER}  != ${OVPN_USER} ]]; then
	report 'WARNING' $ID_OVPN_WRONGPUSER "Wrong process owner (${PUSER}) - required is ${OVPN_USER}"
    fi

    if [[ ${PGROUP} != ${OVPN_GROUP} ]]; then
	report 'WARNING' $ID_OVPN_WRONGPGRP "Wrong process group (${PGROUP}) - required is ${OVPN_GROUP}"
    fi

    unset PUSER
    unset PGROUP

    if [[ -n "${CURRENT_CONF}" ]]; then
	case "${CURRENT_CONF}" in
	    /*)
		CURRENT_CONF=$(rel2abs "${CURRENT_CONF}")
		;;
	    *)
		CURRENT_CONF=$(rel2abs "$(proc_getenv ${PROCDIR}/environ PWD)/${CURRENT_CONF}")
		;;
	esac

	if [[ "${CURRENT_CONF}" != "${DEFAULT_CONF}" ]]; then
	    # check current
	    if [[ -f "${CURRENT_CONF}" ]]; then
		__check_ovpn_config "${PROCDIR}" "${CURRENT_CONF}"
	    fi
	fi
    fi

    # check default
    if [[ -f "${DEFAULT_CONF}" ]]; then
	__check_ovpn_config "${PROCDIR}" "${DEFAULT_CONF}"
    fi
}

########################################

check_env || test_exit ${E_FAIL} "Internal error: check_env failed"

for PID in $(netstat -nltup | sed -n 's|^.*[[:space:]]\+\([0-9]\{1,5\}\)/openvpn|\1|p' | sort | uniq); do
    CHK_openvpn "openvpn" "${PROC}/${PID}/"
done

test_exit ${E_OK}
# EOF
