#!/usr/bin/env python


import selinux as se
import os
import commands as cmd
from python_defs import *



class DiffEntry:
	"""One entry of dictionaries difference image."""
	def __init__(self, key, new = None, old = None):
		self.key = key
		self.old = old
		self.new = new
	def added(self):
		'Entry was added'
		return self.old == None
	def deleted(self):
		'Entry was deleted'
		return self.new == None
	

def dictdiff(old, new):
	"""Returns diff of two dictionaries"""
	diff = {}
	newkeys = new.keys()

	for key in old.keys():
		if new.has_key(key):
			newkeys.remove(key)
			if old[key] != new[key]:
				diff[key] = DiffEntry(key, old[key], new[key])
		else:
			diff[key] = DiffEntry(key, old[key], None)

	for key in newkeys:
		diff[key] = DiffEntry(key, None, new[key])

	return diff

def is_active(v):
	if v in ('yes', 'on', '1', 1):
		return 1
	return 0

def parse_bools(str):
	result = {}
	for line in str.split("\n"):
		if (not line): continue
		k,v = line.split(' --> ')
		result[k] = is_active(v)
	return result

def get_bools():
	bools = se.security_get_boolean_names()[1]
	result = {}
	for name in bools:
		result[name] = se.security_get_boolean_active(name)
	return result

def get_bools_file(fname):
	if (not os.access(fname, os.R_OK)):
		return None
	return parse_bools(open(fname).read())

def save_bools_file(bools, fname):
	f = open(fname, 'w')
	for b,v in bools.items():
		if v: v = 'on'
		else: v = 'off'
		f.write("%s --> %s\n" % (b, v))
	f.close()


# message IDs
ID_FIRSTRUN = 1 << 4
ID_ENABLED  = 2 << 4
ID_IS_ON    = 1
ID_IS_OFF   = 2
ID_IS_NONE  = 3
ID_WAS_ON   = 1 << 2
ID_WAS_OFF  = 2 << 2
ID_WAS_NONE = 3 << 2


messages = {
	ID_FIRSTRUN: "Test is run for first time, skipping",
	ID_IS_ON:    "Boolean %s is now on",
	ID_IS_OFF:   "Boolean %s is now off",
	ID_IS_NONE:  "Boolean %s does not exist now",
	ID_WAS_ON:   "was on before.",
	ID_WAS_OFF:  "was off before.",
	ID_WAS_NONE: "did not exist before."
}

def bool_state_to_id(state, was = False):
	if is_active(state):
		id = ID_IS_ON
	elif state == None:
		id = ID_IS_NONE
	else:
		id = ID_IS_OFF
	if was:
		id = id << 2
	return id


def bools_test():
	old_fname = os.path.join(testenv['TDATA_DIR'], testenv['OLDBOOLS_FILE'])
	bools = get_bools()
	testname = testenv['NAME']
	if (not bools):
		test_exit(E_FAIL, testname, "Unable to retrive booleans information")
	oldbools = get_bools_file(old_fname)
	if (not oldbools):
		save_bools_file(bools, old_fname)
		report(WARNING, ID_FIRSTRUN, "This is a first run of the test. Some parts of audit are skipped.")
		return
	differences = dictdiff(oldbools, bools)
	for bool, diff in differences.items():
		newid = bool_state_to_id(diff.old)
		oldid = bool_state_to_id(diff.new, True)
		msg = messages[newid] + ", %s"
		msg = msg % (bool, messages[oldid])
		report(WARNING, newid | oldid, msg)

def get_msgtype(str):
	if str == 'ERROR'  : return ERROR
	if str == 'WARNING': return WARNING
	if str == 'INFO'   : return INFO
	if str == 'DEBUG'  : return DEBUG
	if str == 'HINT'   : return HINT
	
def enabled_test():
	mode = cmd.getoutput('getenforce')
	if mode != testenv['MODE'] and mode != 'Enforcing':
		report(get_msgtype(testenv['DISABLED_MSGTYPE']), ID_ENABLED, "Selinux is in " + mode + " mode.")
		report(HINT, ID_ENABLED, "Using Enforing mode is highly recommended. See selinux manual page for switching to Enforcing mode.")

	
def main():
	report(INFO, 0, "Starting booleans change test")
	bools_test()
	enabled_test()

if __name__ == '__main__':
	main()
	
